//! Traversible binary trees.

pub mod algorithms;
pub mod unbalanced;

use std::rc::Rc;

use crate::func::*;

/// Result returned by [`Comparator::pick_smaller`].
#[derive(Debug, PartialEq)]
pub enum Comparison {
    L,
    /// Compared values were equal.
    E,
    R,
}

fn and(_l: (), _r: ()) {}

/// Returns [`Comparison`] saying which value is smaller.
///
/// ```rust
/// # use radn_rs::flow::traversible::*;
/// assert_eq!(DefaultComparator.pick_smaller(&1, &3), Comparison::L);
/// assert_eq!(DefaultComparator.pick_smaller(&2, &2), Comparison::E);
/// assert_eq!(DefaultComparator.pick_smaller(&3, &1), Comparison::R);
/// ```
pub trait Comparator<A> {
    fn pick_smaller(&self, kl: &A, kr: &A) -> Comparison;
}

/// Implementation of a [Comparator] relying on [`PartialOrd`].
pub struct DefaultComparator;

impl<A: PartialOrd> Comparator<A> for DefaultComparator {
    fn pick_smaller(&self, kl: &A, kr: &A) -> Comparison {
        if kl < kr {
            Comparison::L
        } else if kr < kl {
            Comparison::R
        } else {
            Comparison::E
        }
    }
}

pub type Split<'a, T, A, D> = (
    Rc<dyn TraversibleBinaryTree<'a, T, A, D>>,
    Rc<dyn TraversibleBinaryTree<'a, T, A, D>>,
    Rc<A>,
);

pub trait TraversibleBinaryNode<'a, T: Monad<'a>, A: 'a, D: 'a + PartialEq>: 'a {
    fn split(&self) -> Split<'a, T, A, D>;

    fn to_tree(self: Rc<Self>) -> Rc<dyn TraversibleBinaryTree<'a, T, A, D>>;
}

pub trait TraversibleBinaryReference<'a, T: Monad<'a>, A: 'a, D: 'a + PartialEq>: 'a {
    fn resolve(&self) -> Wrap<'a, Rc<dyn TraversibleBinaryNode<'a, T, A, D>>, T>;

    /// This should be enough to compare reference for equality.
    fn data(&self) -> D;
}

pub trait TraversibleBinaryTree<'a, T: Monad<'a>, A: 'a, D: 'a + PartialEq>: 'a {
    fn refer(&self) -> Option<Rc<dyn TraversibleBinaryReference<'a, T, A, D>>>;
}
