//! Standard extensions to [`crate::rcore`].

pub mod atomic;
pub mod cast;
pub mod collections;
pub mod fallible;
pub mod inlining;
mod local_origin;
pub mod nullable;
pub mod point;
pub mod tracing;
pub mod typeless;
mod wrapped_origin;

use std::{error::Error, fmt::Display, rc::Rc};

use crate::func::*;
use crate::rcore::*;

impl Display for Address {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}@{}", hex::encode(self.point), self.index))
    }
}

/// Extension trait for [Serializable]s.
pub trait ExtSerializable: Serializable {
    /// Serialize into a [Vec] of bytes.
    fn bytes(&self) -> Vec<u8>;
}

impl<S: Serializable> ExtSerializable for S {
    fn bytes(&self) -> Vec<u8> {
        let mut vec = Vec::new();
        self.serialize(&mut vec);
        vec
    }
}

/// Extention trait for simpler conversion between [`Context::T`] and [`Context::Fallible`].
///
/// Until either Rust type system or [`crate::func`] take serious changes,
/// this is the preferred way to switch between [Wrapped] and [fallible].
pub trait FallibleContext<'a>: Context<'a> {
    /// Convert a fallible wrapped into a wrapped result.
    fn unstuff<A: 'a, E: 'a>(
        wa: WrapE<'a, A, E, Self::Fallible>,
    ) -> Wrap<'a, Result<A, E>, Self::T> {
        Self::Fallible::unstuff(wa)
    }

    /// Convert a wrapped result into a fallible wrapped.
    fn stuff<A: 'a, E: 'a>(fa: Wrap<'a, Result<A, E>, Self::T>) -> WrapE<'a, A, E, Self::Fallible> {
        Self::Fallible::stuff(fa)
    }
}

impl<'a, Ctx: Context<'a>> FallibleContext<'a> for Ctx {}
